@@ -29,11 +29,12 @@ from affine import Affine
 from ._interop import have, is_dask_collection
 from ._rgba import colorize, to_rgba
 from .crs import CRS, CRSError, SomeCRS, norm_crs_or_error
+from .gcp import GCPGeoBox, GCPMapping
 from .geobox import Coordinate, GeoBox
 from .geom import Geometry
 from .math import affine_from_axis
 from .overlap import compute_output_geobox
-from .types import Resolution, resxy_
+from .types import Resolution, resxy_, xy_
 
 # pylint: disable=import-outside-toplevel
 if have.rasterio:
@@ -45,6 +46,7 @@ if have.rasterio:
 XarrayObject = Union[xarray.DataArray, xarray.Dataset]
 XrT = TypeVar("XrT", xarray.DataArray, xarray.Dataset)
 F = TypeVar("F", bound=Callable)
+SomeGeoBox = Union[GeoBox, GCPGeoBox]
 
 _DEFAULT_CRS_COORD_NAME = "spatial_ref"
 
@@ -58,7 +60,8 @@ class GeoState:
     spatial_dims: Optional[Tuple[str, str]] = None
     transform: Optional[Affine] = None
     crs: Optional[CRS] = None
-    geobox: Optional[GeoBox] = None
+    geobox: Optional[SomeGeoBox] = None
+    gcp: Optional[GCPMapping] = None
 
 
 def _get_crs_from_attrs(obj: XarrayObject, sdims: Tuple[str, str]) -> Optional[CRS]:
@@ -147,13 +150,20 @@ def spatial_dims(
     return None
 
 
-def _mk_crs_coord(crs: CRS, name: str = _DEFAULT_CRS_COORD_NAME) -> xarray.DataArray:
+def _mk_crs_coord(
+    crs: CRS,
+    name: str = _DEFAULT_CRS_COORD_NAME,
+    gcps=None,
+) -> xarray.DataArray:
     # pylint: disable=protected-access
 
     cf = crs.proj.to_cf()
     epsg = 0 if crs.epsg is None else crs.epsg
     crs_wkt = cf.get("crs_wkt", None) or crs.wkt
 
+    if gcps is not None:
+        cf["gcps"] = _gcps_to_json(gcps)
+
     return xarray.DataArray(
         numpy.asarray(epsg, "int32"),
         name=name,
@@ -162,6 +172,24 @@ def _mk_crs_coord(crs: CRS, name: str = _DEFAULT_CRS_COORD_NAME) -> xarray.DataA
     )
 
 
+def _gcps_to_json(gcps):
+    def _to_feature(p):
+        coords = [p.x, p.y] if p.z is None else [p.x, p.y, p.z]
+
+        return {
+            "type": "Feature",
+            "properties": {
+                "id": str(p.id),
+                "info": (p.info or ""),
+                "row": p.row,
+                "col": p.col,
+            },
+            "geometry": {"type": "Point", "coordinates": coords},
+        }
+
+    return {"type": "FeatureCollection", "features": list(map(_to_feature, gcps))}
+
+
 def _coord_to_xr(name: str, c: Coordinate, **attrs) -> xarray.DataArray:
     """
     Construct xr.DataArray from named Coordinate object.
@@ -209,7 +237,7 @@ def assign_crs(
 
 
 def xr_coords(
-    gbox: GeoBox, crs_coord_name: Optional[str] = _DEFAULT_CRS_COORD_NAME
+    gbox: SomeGeoBox, crs_coord_name: Optional[str] = _DEFAULT_CRS_COORD_NAME
 ) -> Dict[Hashable, xarray.DataArray]:
     """
     Dictionary of Coordinates in xarray format.
@@ -228,8 +256,16 @@ def xr_coords(
     if crs is not None:
         attrs["crs"] = str(crs)
 
-    if gbox.axis_aligned:
+    gcps = None
+
+    if isinstance(gbox, GCPGeoBox):
         coords: Dict[Hashable, xarray.DataArray] = {
+            name: _mk_pixel_coord(name, sz, None)
+            for name, sz in zip(gbox.dimensions, gbox.shape)
+        }
+        gcps = gbox.gcps()
+    elif gbox.axis_aligned:
+        coords = {
             name: _coord_to_xr(name, coord, **attrs)
             for name, coord in gbox.coordinates.items()
         }
@@ -240,17 +276,22 @@ def xr_coords(
         }
 
     if crs_coord_name is not None and crs is not None:
-        coords[crs_coord_name] = _mk_crs_coord(crs, crs_coord_name)
+        coords[crs_coord_name] = _mk_crs_coord(crs, crs_coord_name, gcps=gcps)
 
     return coords
 
 
-def _mk_pixel_coord(name: str, sz: int, transform: Affine) -> xarray.DataArray:
+def _mk_pixel_coord(
+    name: str,
+    sz: int,
+    transform: Optional[Affine],
+) -> xarray.DataArray:
     data = numpy.arange(0.5, sz, dtype="float32")
     xx = xarray.DataArray(
         data, coords={name: data}, dims=(name,), attrs={"units": "pixel"}
     )
-    xx.encoding["_transform"] = transform[:6]
+    if transform is not None:
+        xx.encoding["_transform"] = transform[:6]
     return xx
 
 
@@ -290,18 +331,31 @@ def _extract_crs(crs_coord: xarray.DataArray) -> Optional[CRS]:
         return None
 
 
-def _locate_geo_info(src: XarrayObject) -> GeoState:
-    sdims = spatial_dims(src, relaxed=True)
-    if sdims is None:
-        return GeoState()
+def _extract_gcps(crs_coord: xarray.DataArray) -> Optional[GCPMapping]:
+    gcps = crs_coord.attrs.get("gcps", None)
+    if gcps is None:
+        return None
+    crs = _extract_crs(crs_coord)
+    try:
+        wld = Geometry(gcps, crs=crs)
+        pix = [
+            xy_(f["properties"]["col"], f["properties"]["row"])
+            for f in gcps["features"]
+        ]
+        return GCPMapping(pix, wld)
+    except (IndexError, KeyError, ValueError):
+        return None
 
-    transform: Optional[Affine] = None
-    crs: Optional[CRS] = None
-    geobox: Optional[GeoBox] = None
-    fallback_res: Optional[Resolution] = None
+
+def _extract_transform(src: XarrayObject, sdims: Tuple[str, str]) -> Optional[Affine]:
+    if any(dim not in src.coords for dim in sdims):
+        # special case of no spatial dims at all
+        # happens for GCP sources loaded by rioxarray
+        return None
 
     _yy, _xx = (src[dim] for dim in sdims)
     rx, ry = (coord.attrs.get("resolution", None) for coord in (_xx, _yy))
+    fallback_res = None
     if rx is not None and ry is not None:
         fallback_res = resxy_(float(rx), float(ry))
 
@@ -309,31 +363,49 @@ def _locate_geo_info(src: XarrayObject) -> GeoState:
         transform = affine_from_axis(_xx.values, _yy.values, fallback_res)
     except ValueError:
         # this can fail when any dimension is shorter than 2 elements
-        pass
+        return None
+
+    _pix2world = _xx.encoding.get("_transform", None)
+    if _pix2world is not None:
+        # non-axis aligned geobox detected
+        # adjust transform
+        #  world <- pix' <- pix
+        transform = Affine(*_pix2world) * transform
+    return transform
+
+
+def _locate_geo_info(src: XarrayObject) -> GeoState:
+    # pylint: disable=too-many-locals
+    sdims = spatial_dims(src, relaxed=True)
+    if sdims is None:
+        return GeoState()
+
+    transform = _extract_transform(src, sdims)
+    crs: Optional[CRS] = None
+    geobox: Optional[SomeGeoBox] = None
+    gcp: Optional[GCPMapping] = None
+
+    ny, nx = (src.coords[dim].shape[0] for dim in sdims)
 
     _crs_coords = _locate_crs_coords(src)
-    num_candiates = len(_crs_coords)
-    if num_candiates > 0:
-        if num_candiates > 1:
+    num_candidates = len(_crs_coords)
+    if num_candidates > 0:
+        if num_candidates > 1:
             warnings.warn("Multiple CRS coordinates are present")
         crs = _extract_crs(_crs_coords[0])
+        gcp = _extract_gcps(_crs_coords[0])
     else:
         # try looking in attributes
         crs = _get_crs_from_attrs(src, sdims)
 
-    if transform is not None:
-        _pix2world = _xx.encoding.get("_transform", None)
-        if _pix2world is not None:
-            # non-axis aligned geobox detected
-            # adjust transform
-            #  world <- pix' <- pix
-            transform = Affine(*_pix2world) * transform
-
-        nx = _xx.shape[0]
-        ny = _yy.shape[0]
+    if gcp is not None:
+        geobox = GCPGeoBox((ny, nx), gcp, transform)
+    elif transform is not None:
         geobox = GeoBox((ny, nx), transform, crs)
 
-    return GeoState(spatial_dims=sdims, transform=transform, crs=crs, geobox=geobox)
+    return GeoState(
+        spatial_dims=sdims, transform=transform, crs=crs, geobox=geobox, gcp=gcp
+    )
 
 
 def _wrap_op(method: F) -> F:
@@ -347,6 +419,64 @@ def _wrap_op(method: F) -> F:
 
 
 def xr_reproject(
+    src: XrT,
+    how: Union[SomeCRS, GeoBox],
+    *,
+    resampling: Union[str, int] = "nearest",
+    dst_nodata: Optional[float] = None,
+    **kw,
+) -> XrT:
+    """
+    Reproject raster to different projection/resolution.
+
+    This method uses :py:mod:`rasterio`.
+    """
+    if isinstance(src, xarray.DataArray):
+        return _xr_reproject_da(
+            src, how, resampling=resampling, dst_nodata=dst_nodata, **kw
+        )
+    return _xr_reproject_ds(
+        src, how, resampling=resampling, dst_nodata=dst_nodata, **kw
+    )
+
+
+def _xr_reproject_ds(
+    src: Any,
+    how: Union[SomeCRS, GeoBox],
+    *,
+    resampling: Union[str, int] = "nearest",
+    dst_nodata: Optional[float] = None,
+    **kw,
+) -> xarray.Dataset:
+    assert isinstance(src, xarray.Dataset)
+
+    if have.rasterio is False:  # pragma: nocover
+        raise RuntimeError("Please install `rasterio` to use this method")
+
+    assert isinstance(src.odc, ODCExtensionDs)
+    if src.odc.geobox is None:
+        raise ValueError("Can not reproject non-georegistered array.")
+
+    if isinstance(how, GeoBox):
+        dst_geobox = how
+    else:
+        dst_geobox = src.odc.output_geobox(how)
+
+    def _maybe_reproject(dv: xarray.DataArray):
+        if dv.odc.geobox is None:
+            # pass-through data variables without a geobox
+            strip_coords = [str(c.name) for c in _locate_crs_coords(dv)]
+            if len(strip_coords) > 0:
+                dv = dv.drop_vars(strip_coords)
+            return dv
+        return _xr_reproject_da(
+            dv, how=dst_geobox, resampling=resampling, dst_nodata=dst_nodata, **kw
+        )
+
+    return src.map(_maybe_reproject)
+
+
+def _xr_reproject_da(
     src: Any,
     how: Union[SomeCRS, GeoBox],
     *,
@@ -361,10 +491,8 @@ def xr_reproject(
     """
     assert isinstance(src, xarray.DataArray)
 
-    if have.rasterio is False:
-        raise RuntimeError(
-            "Please install `rasterio` to use this method"
-        )  # pragma: nocover
+    if have.rasterio is False:  # pragma: nocover
+        raise RuntimeError("Please install `rasterio` to use this method")
 
     assert isinstance(src.odc, ODCExtensionDa)  # for mypy sake
     src_gbox = src.odc.geobox
@@ -388,7 +516,7 @@ def xr_reproject(
     dst = numpy.empty(dst_shape, dtype=src.dtype)
     src_nodata = kw.pop("src_nodata", None)
     if src_nodata is None:
-        src_nodata = src.attrs.get("nodata", None)
+        src_nodata = src.odc.nodata
     if dst_nodata is None:
         dst_nodata = src_nodata
 
@@ -407,6 +535,7 @@ def xr_reproject(
     attrs = src.attrs.copy()
     if dst_nodata is None:
         attrs.pop("nodata", None)
+        attrs.pop("_FillValue", None)
     else:
         attrs.update(nodata=dst_nodata)
 
@@ -446,8 +575,8 @@ class ODCExtension:
         return self._state.crs
 
     @property
-    def geobox(self) -> Optional[GeoBox]:
-        """Query :py:class:`~odc.geo.geobox.GeoBox`."""
+    def geobox(self) -> Optional[SomeGeoBox]:
+        """Query :py:class:`~odc.geo.geobox.GeoBox` or :py:class:`~odc.geo.gcp.GCPGeoBox`."""
         return self._state.geobox
 
     def output_geobox(self, crs: SomeCRS, **kw) -> GeoBox:
@@ -505,13 +634,24 @@ class ODCExtensionDa(ODCExtension):
         """See :py:meth:`odc.geo.xr.assign_crs`."""
         return assign_crs(self._xx, crs=crs, crs_coord_name=crs_coord_name)
 
+    @property
+    def nodata(self) -> Optional[float]:
+        """Extract ``nodata/_FillValue`` attribute if set."""
+        attrs = self._xx.attrs
+        for k in ["nodata", "_FillValue"]:
+            nodata = attrs.get(k, None)
+            if nodata is not None:
+                return float(nodata)
+
+        return None
+
     colorize = _wrap_op(colorize)
 
     if have.rasterio:
         write_cog = _wrap_op(write_cog)
         to_cog = _wrap_op(to_cog)
         compress = _wrap_op(compress)
-        reproject = _wrap_op(xr_reproject)
+        reproject = _wrap_op(_xr_reproject_da)
         add_to = _wrap_op(add_to)
 
 
@@ -523,16 +663,16 @@ class ODCExtensionDs(ODCExtension):
 
     def __init__(self, ds: xarray.Dataset):
         ODCExtension.__init__(self, _locate_geo_info(ds))
-        self._ds = ds
+        self._xx = ds
 
     @property
     def uncached(self) -> "ODCExtensionDs":
-        return ODCExtensionDs(self._ds)
+        return ODCExtensionDs(self._xx)
 
     def assign_crs(
         self, crs: SomeCRS, crs_coord_name: str = _DEFAULT_CRS_COORD_NAME
     ) -> xarray.Dataset:
-        return assign_crs(self._ds, crs=crs, crs_coord_name=crs_coord_name)
+        return assign_crs(self._xx, crs=crs, crs_coord_name=crs_coord_name)
 
     def to_rgba(
         self,
@@ -541,7 +681,10 @@ class ODCExtensionDs(ODCExtension):
         vmin: Optional[float] = None,
         vmax: Optional[float] = None,
     ) -> xarray.DataArray:
-        return to_rgba(self._ds, bands=bands, vmin=vmin, vmax=vmax)
+        return to_rgba(self._xx, bands=bands, vmin=vmin, vmax=vmax)
+
+    if have.rasterio:
+        reproject = _wrap_op(_xr_reproject_ds)
 
 
 ODCExtensionDs.to_rgba.__doc__ = to_rgba.__doc__
@@ -567,7 +710,7 @@ def register_geobox():
 
 def wrap_xr(
     im: Any,
-    gbox: GeoBox,
+    gbox: SomeGeoBox,
     *,
     time=None,
     nodata=None,
@@ -610,7 +753,7 @@ def wrap_xr(
 
 
 def xr_zeros(
-    geobox: GeoBox,
+    geobox: SomeGeoBox,
     dtype="float64",
     *,
     chunks: Optional[Union[Tuple[int, int], Tuple[int, int, int]]] = None,
@@ -686,10 +829,8 @@ def rasterize(
     """
     # pylint: disable=import-outside-toplevel
 
-    if have.rasterio is False:
-        raise RuntimeError(
-            "Please install `rasterio` to use this method"
-        )  # pragma: nocover
+    if have.rasterio is False:  # pragma: nocover
+        raise RuntimeError("Please install `rasterio` to use this method")
 
     from rasterio.features import geometry_mask
 
