@@ -23,11 +23,13 @@ from typing import (
 
 import numpy
 from affine import Affine
+from pyproj.aoi import AreaOfInterest
 from shapely import geometry, ops
 from shapely.geometry import base
 
 from .crs import CRS, CRSMismatchError, MaybeCRS, SomeCRS, norm_crs, norm_crs_or_error
-from .types import SomeShape, SupportsCoords, shape_
+from .math import edge_index, quasi_random_r2
+from .types import SomeShape, SupportsCoords, Unset, shape_
 
 CoordList = List[Tuple[float, float]]
 
@@ -97,6 +99,12 @@ class BoundingBox(Sequence[float]):
     def __str__(self) -> str:
         return self.__repr__()
 
+    def __and__(self, other: "BoundingBox") -> "BoundingBox":
+        return bbox_intersection([self, other])
+
+    def __or__(self, other: "BoundingBox") -> "BoundingBox":
+        return bbox_union([self, other])
+
     def buffered(self, xbuff: float, ybuff: Optional[float] = None) -> "BoundingBox":
         """
         Return a new BoundingBox, buffered in the x and y dimensions.
@@ -240,6 +248,89 @@ class BoundingBox(Sequence[float]):
         p2 = transform * shape_(shape).xy
         return BoundingBox.from_points(p1, p2, crs=crs)
 
+    @property
+    def aoi(self) -> AreaOfInterest:
+        """
+        Interop with pyproj.
+        """
+        if self._crs is None or self._crs == "epsg:4326":
+            return AreaOfInterest(*self._box)
+        return AreaOfInterest(*self.to_crs("epsg:4326").bbox)
+
+    def round(self) -> "BoundingBox":
+        """
+        Expand bounding box to nearest integer on all sides.
+        """
+        x0, y0, x1, y1 = self._box
+        return BoundingBox(
+            math.floor(x0), math.floor(y0), math.ceil(x1), math.ceil(y1), crs=self._crs
+        )
+
+    def boundary(self, pts_per_side: int = 2) -> "Geometry":
+        """
+        Points along the perimeter as a linear ring.
+        """
+        x0, y0, x1, y1 = self._box
+        xx = numpy.linspace(x0, x1, pts_per_side, dtype="float32")
+        yy = numpy.linspace(y0, y1, pts_per_side, dtype="float32")
+        return line(
+            [
+                (float(xx[ix]), float(yy[iy]))
+                for iy, ix in edge_index((pts_per_side, pts_per_side), closed=True)
+            ],
+            self.crs,
+        )
+
+    def qr2sample(
+        self,
+        n: int,
+        padding: Optional[float] = None,
+        with_edges: bool = False,
+        offset: int = 0,
+    ) -> "Geometry":
+        """
+        Generate quasi-random sample of points within this boundingbox.
+
+        :param n:
+           Number of points
+
+        :param padding:
+           Minimal distance to the edge
+
+        :param offset:
+           Offset into quasi-random sequence from where to start
+
+        :param edges:
+           Also include samples along the edge and corners
+
+        References: http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
+        """
+        # pylint: disable=too-many-locals
+        x0, y0, x1, y1 = self._box
+        nx = x1 - x0
+        ny = y1 - y0
+        pts = quasi_random_r2(n, offset=offset)
+        s = numpy.asarray([nx, ny], dtype="float32")
+        edge_pts = []
+
+        if with_edges:
+            sample_density = numpy.sqrt(n / (nx * ny))
+            n_side = int(numpy.round(sample_density * min(nx, ny))) + 1
+            n_side = max(2, n_side)
+            edge_pts = self.boundary(n_side).coords[:-1]
+            if padding is None:
+                padding = 0.3 * min(nx, ny) / (n_side - 1)
+
+        if padding is None:
+            pts = pts * s
+        else:
+            pts = pts * (s - 2 * padding) + padding
+
+        pts[:, 0] += x0
+        pts[:, 1] += y0
+
+        return multipoint(pts.tolist() + edge_pts, self.crs)
+
 
 def wrap_shapely(method):
     """
@@ -283,6 +374,29 @@ def force_2d(geojson: Dict[str, Any]) -> Dict[str, Any]:
     return {"type": geojson["type"], "coordinates": go(geojson["coordinates"])}
 
 
+def _geojson_to_shapely(xx: Any) -> base.BaseGeometry:
+    _type = xx.get("type", None)
+
+    def to_geom(x) -> base.BaseGeometry:
+        return geometry.shape(force_2d(x))
+
+    if _type is None:
+        raise ValueError("Not a valid GeoJSON")
+
+    _type = _type.lower()
+    if _type == "featurecollection":
+        features = xx.get("features", [])
+        if len(features) == 1:
+            return to_geom(features[0]["geometry"])
+
+        return _multigeom([to_geom(feature["geometry"]) for feature in features])
+
+    if _type == "feature":
+        return to_geom(xx["geometry"])
+
+    return to_geom(xx)
+
+
 def densify(coords: CoordList, resolution: float) -> CoordList:
     """
     Adds points so they are at most `resolution` units apart.
@@ -335,12 +449,19 @@ class Geometry(SupportsCoords[float]):
             self.geom: base.BaseGeometry = _clone_shapely_geom(geom.geom)
             return
 
+        if crs is None:
+            if isinstance(geom, dict) and geom.get("type", "").lower().startswith(
+                "feature"
+            ):
+                # Assume 4326 for geojson inputs
+                crs = "epsg:4326"
+
         crs = norm_crs(crs)
         self.crs = crs
         if isinstance(geom, base.BaseGeometry):
             self.geom = geom
         elif isinstance(geom, dict):
-            self.geom = geometry.shape(force_2d(geom))
+            self.geom = _geojson_to_shapely(geom)
         else:
             raise ValueError(f"Unexpected type {type(geom)}")
 
@@ -379,8 +500,6 @@ class Geometry(SupportsCoords[float]):
     @wrap_shapely
     def __sub__(self, other: "Geometry") -> "Geometry": return self.__sub__(other)
 
-    def svg(self, *args, **kw) -> str: return self.geom.svg(*args, *kw)
-
     def _repr_svg_(self) -> str: return self.geom._repr_svg_()
 
     @property
@@ -470,7 +589,7 @@ class Geometry(SupportsCoords[float]):
             self.geom.simplify(tolerance, preserve_topology=preserve_topology), self.crs
         )
 
-    def transform(self, func) -> "Geometry":
+    def transform(self, func, *, crs: MaybeCRS = Unset()) -> "Geometry":
         """
         Map through arbitrary transform.
 
@@ -480,6 +599,9 @@ class Geometry(SupportsCoords[float]):
         ``func`` maps ``x, y``, and optionally ``z`` to output ``xp, yp, zp``. The input parameters
         may be iterable types like lists or arrays or single values. The output shall be of the same
         type: scalars in, scalars out; lists in, lists out.
+
+        :param crs: If supplied override output CRS with that value, ``crs=None`` will remove CRS
+        for the output, default is to keep the original projection.
         """
         if isinstance(func, Affine):
 
@@ -491,7 +613,10 @@ class Geometry(SupportsCoords[float]):
         else:
             _geom = ops.transform(func, self.geom)
 
-        return Geometry(_geom, self.crs)
+        if isinstance(crs, Unset):
+            crs = self.crs
+
+        return Geometry(_geom, crs)
 
     def _to_crs(self, crs: CRS) -> "Geometry":
         assert self.crs is not None
@@ -528,7 +653,7 @@ class Geometry(SupportsCoords[float]):
            geographic projections. Currently only works in few specific cases (source CRS is smooth
            over the dateline).
         """
-        crs = norm_crs_or_error(crs)
+        crs = norm_crs_or_error(crs, self)
         if self.crs == crs:
             return self
 
@@ -550,6 +675,12 @@ class Geometry(SupportsCoords[float]):
 
         return geom._to_crs(crs)
 
+    def assign_crs(self, crs: MaybeCRS) -> "Geometry":
+        """
+        Same geometry but with crs changed.
+        """
+        return Geometry(self.geom, crs=crs)
+
     def geojson(
         self,
         properties: Optional[Dict[str, Any]] = None,
@@ -580,7 +711,28 @@ class Geometry(SupportsCoords[float]):
 
         :return: GeoJSON Feature dictionary
         """
-        gg = self.to_crs("epsg:4326", resolution=resolution, wrapdateline=wrapdateline)
+        if self.type == "GeometryCollection":
+            return {
+                "type": "FeatureCollection",
+                "features": [
+                    g.geojson(
+                        properties=properties,
+                        simplify=simplify,
+                        resolution=resolution,
+                        wrapdateline=wrapdateline,
+                        **props,
+                    )
+                    for g in self.geoms
+                ],
+            }
+
+        if self.crs is not None:
+            gg = self.to_crs(
+                "epsg:4326", resolution=resolution, wrapdateline=wrapdateline
+            )
+        else:
+            gg = self
+
         if simplify > 0:
             gg = gg.simplify(simplify)
         if properties is None:
@@ -649,6 +801,23 @@ class Geometry(SupportsCoords[float]):
         """
         return self.transform(A)
 
+    def svg(self, *args, **kw) -> str:
+        """
+        Returns SVG path element (wraps shapely).
+
+        This method will drop parameters not supported by the underlying shapely geometry.
+
+        :param scale_factor: Multiplication factor for the SVG stroke-width.  Default is 1
+        :param stroke_color: Color for lines and rings
+        :param fill_color: Color for poinits and polygons
+        :param opacity: Float number between 0 and 1 for color opacity. Default value is 0.6
+        """
+        from inspect import signature  # pylint: disable=import-outside-toplevel
+
+        valid_params = signature(self.geom.svg).parameters
+        opts = {k: v for k, v in kw.items() if k in valid_params}
+        return self.geom.svg(*args, **opts)
+
     def svg_path(self, ndecimal: int = 6) -> str:
         """
         Produce SVG path text.
@@ -682,9 +851,15 @@ class Geometry(SupportsCoords[float]):
         """
 
         if self.type == "Polygon":
-            # TODO: deal with interior rings
             pts = [(x, y) for x, y in self.exterior.points if pred(x, y)]
-            return polygon(pts, self.crs)
+            # prune inner polygon points
+            inners = [
+                [(x, y) for x, y in ring.points if pred(x, y)]
+                for ring in self.interiors
+            ]
+            # prune inner polygons that don't have enough points
+            inners = [ring for ring in inners if len(ring) >= 3]
+            return polygon(pts, self.crs, *inners)
 
         if self.type in ("LinearRing", "LineString"):
             pts = [(x, y) for x, y in self.points if pred(x, y)]
@@ -749,7 +924,7 @@ def projected_lon(
     :param lat: Optionally limit range of the line
     :param step: Line "resolution" in degrees
     """
-    crs = norm_crs_or_error(crs)
+    crs = norm_crs_or_error(crs, lon)
     yy = numpy.arange(lat[0], lat[1], step, dtype="float32")
     xx = numpy.full_like(yy, lon)
     tr = CRS("EPSG:4326").transformer_to_crs(crs)
@@ -942,25 +1117,26 @@ def sides(poly: Geometry) -> Iterable[Geometry]:
         yield line([p1, p2], crs)
 
 
-def multigeom(geoms: Iterable[Geometry]) -> Geometry:
-    """Construct ``Multi{Polygon|LineString|Point}``."""
-    geoms = list(geoms)  # force into list
+def _multigeom(geoms: List[base.BaseGeometry]) -> base.BaseGeometry:
     src_type = {g.type for g in geoms}
-    raw_geoms = [g.geom for g in geoms]
-    crs = common_crs(geoms)  # will raise if some differ
-
     if len(src_type) > 1:
-        return Geometry(geometry.GeometryCollection(raw_geoms), crs)
+        return geometry.GeometryCollection(geoms)
 
     src_type = src_type.pop()
     if src_type == "Polygon":
-        return Geometry(geometry.MultiPolygon(raw_geoms), crs)
+        return geometry.MultiPolygon(geoms)
     if src_type == "Point":
-        return Geometry(geometry.MultiPoint(raw_geoms), crs)
+        return geometry.MultiPoint(geoms)
     if src_type == "LineString":
-        return Geometry(geometry.MultiLineString(raw_geoms), crs)
+        return geometry.MultiLineString(geoms)
+    return geometry.GeometryCollection(geoms)
 
-    raise ValueError("Only understand Polygon|LineString|Point")
+
+def multigeom(geoms: Iterable[Geometry]) -> Geometry:
+    """Construct ``Multi{Polygon|LineString|Point}``."""
+    geoms = list(geoms)  # force into list
+    crs = common_crs(geoms)  # will raise if some differ
+    return Geometry(_multigeom([g.geom for g in geoms]), crs)
 
 
 ###########################################
@@ -992,6 +1168,27 @@ def unary_intersection(geoms: Iterable[Geometry]) -> Geometry:
     return functools.reduce(Geometry.intersection, geoms)
 
 
+def triangulate(pts: Geometry, **kw) -> Geometry:
+    """
+     Creates the Delaunay triangulation and returns a geometry collection.
+
+     The source may be any geometry type. All vertices of the geometry will be
+     used as the points of the triangulation.
+
+    :param tolerance:
+        From the GEOS documentation:
+        tolerance is the snapping tolerance used to improve the robustness of
+        the triangulation computation. A tolerance of 0.0 specifies that no
+        snapping will take place.
+
+     :param edges:
+        If edges is False, a geometry collection of Polygons (triangles) will be returned (default).
+        Otherwise LineString edges are returned.
+    """
+    geom = geometry.GeometryCollection(ops.triangulate(pts.geom, **kw))
+    return Geometry(geom, pts.crs)
+
+
 def intersects(a: Geometry, b: Geometry) -> bool:
     """
     Check for intersection.
