@@ -8,12 +8,23 @@ Various mathy helpers.
 Minimal dependencies in this module.
 """
 from math import ceil, floor, fmod, isfinite
-from typing import List, Literal, Optional, Sequence, Tuple, TypeVar, Union
+from typing import (
+    Any,
+    Iterator,
+    List,
+    Literal,
+    Optional,
+    Sequence,
+    Tuple,
+    TypeVar,
+    Union,
+)
 
 import numpy as np
 from affine import Affine
+from numpy.polynomial.polynomial import polygrid2d, polyval2d
 
-from .types import XY, Resolution, SomeResolution, res_, resxy_, xy_
+from .types import XY, Resolution, SomeResolution, SomeShape, res_, resxy_, shape_, xy_
 
 AffineX = TypeVar("AffineX", np.ndarray, Affine)
 
@@ -364,7 +375,7 @@ def decompose_rws(A: AffineX) -> Tuple[AffineX, AffineX, AffineX]:
        * ``(R*(-I))*((-I)*S) == R*S``
 
 
-     :return: Rotation, Sheer, Scale ``2x2`` matrices
+     :return: Rotation, Shear, Scale ``2x2`` matrices
     """
     # pylint: disable=too-many-locals
 
@@ -407,6 +418,31 @@ def unstack_xy(pts: np.ndarray) -> List[XY[float]]:
     return [xy_(pt) for pt in pts]
 
 
+def norm_xy(
+    pts: np.ndarray, out: Optional[np.ndarray] = None
+) -> Tuple[np.ndarray, Affine]:
+    """
+    Normalize ``Nx2`` points.
+
+    Scale and translate such that mean is at 0 and mean distance from 0 is sqrt(2).
+
+    :returns: Normalized array and affine matrix of the normalization.
+    """
+    assert pts.ndim == 2 and pts.shape[1] == 2
+    assert out is None or out.shape == pts.shape
+
+    _mean = pts.mean(axis=0)
+    XX = np.subtract(pts, _mean, out=out)
+
+    sx = (((XX**2).sum(axis=1) * 0.5) ** -0.5).mean()
+    XX *= sx
+
+    tx, ty = -_mean * sx
+    A = Affine(sx, 0, tx, 0, sx, ty)
+
+    return XX, A
+
+
 def affine_from_pts(X: Sequence[XY[float]], Y: Sequence[XY[float]]) -> Affine:
     """
     Given points ``X,Y`` compute ``A``, such that: ``Y = A*X``.
@@ -444,6 +480,66 @@ def resolution_from_affine(A: Affine) -> Resolution:
     return resxy_(rx, ry)
 
 
+def edge_index(shape: SomeShape, closed: bool = False) -> Iterator[Tuple[int, int]]:
+    """
+    Like ``ndindex`` in numpy but for edge locations and limited to 2d.
+
+    Returns a sequence of indexes along the edge of a 2d array of a given shape.
+    Order is fixed, starting from ``(0, 0), (0, 1)...``. See example below
+
+    .. code-block::
+
+       # for shape=(10, 8)
+       (0, 0), (0, 1) ... (0, 7),
+       (1, 7), (2, 7) ... (9, 7),
+       (9, 6), (9, 5) ... (9, 0),
+       (8, 0), (7, 0) ... (1, 0),
+       (0, 0) # Back to (0, 0) if closed=True
+
+    """
+    nx, ny = shape_(shape).xy
+    ix, iy = 0, 0
+    for ix in range(nx):
+        yield (0, ix)
+    for iy in range(1, ny):
+        yield (iy, ix)
+    for ix in range(ix - 1, -1, -1):
+        yield (iy, ix)
+    for iy in range(iy - 1, 0, -1):
+        yield (iy, ix)
+    if closed:
+        yield (0, 0)
+
+
+def quasi_random_r2(
+    n: int,
+    shape: Optional[SomeShape] = None,
+    offset: int = 0,
+) -> np.ndarray:
+    """
+    Generate quasi-random set of points.
+
+    Returns quasi-random points in ``[0, 1)`` range. If ``shape=(ny, nx)`` is
+    supplied, points are scaled to ``[0, nx), [0, ny)`` range.
+
+    :param n: Number of points to generate
+    :param offset: Generate from that offset in the sequence
+    :returns: ``nx2`` numpy array
+
+    References: http://extremelearning.com.au/unreasonable-effectiveness-of-quasirandom-sequences/
+    """
+    idx = np.arange(offset, offset + n, dtype="float32")
+    aa = (0.7548776662466927, 0.5698402909980532)
+    xy = np.fmod(np.outer(idx, aa), 1)
+
+    if shape is not None:
+        nx, ny = shape_(shape).wh
+        xy[:, 0] *= nx
+        xy[:, 1] *= ny
+
+    return xy
+
+
 class Bin1D:
     """
     Class for translating continous coordinates to bin index.
@@ -514,3 +610,163 @@ class Bin1D:
         origin = x0 - sz * idx * direction
 
         return Bin1D(sz, origin, direction=direction)
+
+
+class Poly2d:
+    """
+    Wrapper around polyval2d with input normalization.
+    """
+
+    def __init__(self, cc: np.ndarray, A: Affine) -> None:
+        assert cc.shape in [(3, 3, 2), (2, 2, 2)]
+        tol = 1e-6
+        self._cc = cc
+        self._A = A
+        self._safe_to_grid = False
+
+        sx, zx, tx, zy, sy, ty, *_ = A
+        if abs(zx) < tol and abs(zy) < tol:
+            self._norm = lambda x, y: (np.polyval([sx, tx], x), np.polyval([sy, ty], y))
+            self._safe_to_grid = True
+        else:
+            self._norm = lambda x, y: A * (x, y)
+
+    def __call__(self, x: Any, y: Any = None) -> Any:
+        """
+        Evaluate at points (x, y).
+
+        When y is None, x is assumed to be Nx2 array.
+        """
+        if y is None:
+            return self.__call__(x[..., 0], x[..., 1]).T
+        x, y = self._norm(x, y)
+        return polyval2d(x, y, self._cc)
+
+    def grid2d(self, x: Any, y: Any) -> Any:
+        """
+        Evaluate on the Cartesian product of x and y.
+        """
+        assert self._safe_to_grid
+
+        x, y = self._norm(x, y)
+        return polygrid2d(x, y, self._cc)
+
+    def with_input_transform(self, A: Affine) -> "Poly2d":
+        """
+        Make a new transform by chaining a linear mapping ``A`` on input side.
+
+        usefull when cropping original image with GCPs.
+        """
+        return Poly2d(self._cc, self._A * A)
+
+    @staticmethod
+    def fit(aa: np.ndarray, bb: np.ndarray) -> "Poly2d":
+        """
+        Fit 2d polynomial that minimizes ``poly(aa) - bb``.
+
+        Where ``aa``,``bb`` are point correspondences of ``Nx2`` shape.
+        """
+        assert aa.shape[1] == 2
+        assert aa.shape == bb.shape
+        N = aa.shape[0]
+        if N < 3:
+            raise ValueError(f"Need at least 3 points, got {N}")
+
+        aa_, Ain = norm_xy(aa)
+        bb_, Ab = norm_xy(bb)
+
+        if N >= 9:
+            return Poly2d._fit9(aa_, Ain, bb_, Ab)
+
+        if N >= 4:
+            return Poly2d._fit4(aa_, Ain, bb_, Ab)
+
+        return Poly2d._fit3(aa_, Ain, bb_, Ab)
+
+    @staticmethod
+    def _fit9(aa: np.ndarray, Ain: Affine, bb: np.ndarray, Ab: Affine) -> "Poly2d":
+        N = aa.shape[0]
+        assert N >= 9
+        x, y = aa.T
+        AA = np.empty((N, 9), dtype="float64")
+
+        #   1,     y,     y^2
+        #   x,   x*y,   x*y^2
+        # x^2, x^2*y, x^2*y^2
+
+        AA[:, 0] = 1
+        AA[:, 1] = y
+        AA[:, 2] = y * y
+
+        AA[:, 3] = x
+        AA[:, 4] = x * y
+        AA[:, 5] = AA[:, 4] * y
+
+        AA[:, 6] = x * x
+        AA[:, 7] = AA[:, 6] * y
+        AA[:, 8] = AA[:, 7] * y
+
+        cc, *_ = np.linalg.lstsq(AA, bb, rcond=-1)
+
+        # denorm output side, assumes `sx==sy`
+        s, _, tx, _, _, ty, *_ = ~Ab
+        cc = cc * s
+        cc[0, :2] += (tx, ty)
+
+        cc = cc.reshape(3, 3, 2)
+
+        return Poly2d(cc, Ain)
+
+    @staticmethod
+    def _fit4(aa: np.ndarray, Ain: Affine, bb: np.ndarray, Ab: Affine) -> "Poly2d":
+        N = aa.shape[0]
+        assert N >= 4
+        x, y = aa.T
+        AA = np.empty((N, 4), dtype="float64")
+
+        #   1,     y
+        #   x,   x*y
+
+        AA[:, 0] = 1
+        AA[:, 1] = y
+
+        AA[:, 2] = x
+        AA[:, 3] = x * y
+
+        cc, *_ = np.linalg.lstsq(AA, bb, rcond=-1)
+        assert cc.shape == (4, 2)
+
+        # denorm output side, assumes `sx==sy`
+        s, _, tx, _, _, ty, *_ = ~Ab
+        cc = cc * s
+        cc[0, :2] += (tx, ty)
+
+        cc = cc.reshape(2, 2, 2)
+
+        return Poly2d(cc, Ain)
+
+    @staticmethod
+    def _fit3(aa: np.ndarray, Ain: Affine, bb: np.ndarray, Ab: Affine) -> "Poly2d":
+        N = aa.shape[0]
+        assert N >= 3
+        x, y = aa.T
+        AA = np.empty((N, 3), dtype="float64")
+
+        #   1,   y
+        #   x,   _
+
+        # 1, y, x
+        AA[:, 0] = 1
+        AA[:, 1] = y
+        AA[:, 2] = x
+
+        cc, *_ = np.linalg.lstsq(AA, bb, rcond=-1)
+        assert cc.shape == (3, 2)
+
+        # denorm output side, assumes `sx==sy`
+        s, _, tx, _, _, ty, *_ = ~Ab
+        cc = cc * s
+        cc[0, :2] += (tx, ty)
+        cc = np.vstack([cc, np.asarray([0, 0])]).reshape(2, 2, 2)
+
+        return Poly2d(cc, Ain)
