@@ -23,6 +23,16 @@ T1 = TypeVar("T1")
 T2 = TypeVar("T2")
 
 
+class Unset:
+    """
+    Marker for unset values.
+
+    Used where ``None`` can be a valid value.
+    """
+
+    __slots__ = ()
+
+
 class XY(Generic[T]):
     """
     Immutable container for anything X/Y.
@@ -219,7 +229,6 @@ class SupportsCoords(Protocol[T]):
     Needed for Point geometry -> XY conversion.
     """
 
-    # pylint: disable=too-few-public-methods
     @property
     def coords(self) -> List[Tuple[T, T]]:
         ...
@@ -434,3 +443,7 @@ ROI = Tuple[SomeSlice, SomeSlice]
 
 NormalizedROI = Tuple[NormalizedSlice, NormalizedSlice]
 """Normalized 2d slice into an image plane."""
+
+OutlineMode = Union[
+    Literal["native"], Literal["pixel"], Literal["geo"], Literal["auto"]
+]
